@@ -13,6 +13,7 @@
 #include "headers.h"
 #include "matrix.h"
 
+#include "tree.h"
 
 
 #ifdef MPI2
@@ -34,6 +35,7 @@ extern int int_VERBOSE;
 extern int int_TAYLOR;
 extern int int_WEIGHTS;
 extern int int_LARGE_SF;
+extern int int_TREE;
 
 /*
 int int_LARGE_SF; 
@@ -122,6 +124,9 @@ extern double *vector_scale_factor_dep_met_univar_lag_extern;
 extern double y_min_extern;
 extern double y_max_extern;
 
+// tree
+extern KDT * kdt_extern;
+
 int kernel_convolution_weighted_sum(
 int KERNEL_reg,
 int KERNEL_unordered_reg,
@@ -783,6 +788,28 @@ double (* const allck[])(double) = { np_gauss2, np_gauss4, np_gauss6, np_gauss8,
 double (* const allok[])(double, double, double) = { np_owang_van_ryzin, np_oli_racine };
 double (* const alluk[])(int, double, int) = { np_uaa, np_uli_racine };
 
+// in cksup we define a scale length for all kernels, outside of which the kernel is 0
+// for fixed bandwidths and finite support kernels, only points within += 1 scale length
+// contribute at a given point of evaluation
+// when constructing the search box, in 1D the size of the box is
+// x_eval + [-RIGHT_SUPPORT,-LEFT_SUPPORT], the interval is reversed and negated because you want to know if 
+// the evaluation point is in the support of other points, not vice versa :)
+
+// as it stands, CDF's are only partially accelerated by the tree, but that is fixable
+#define SQRT5 2.23606797749979
+#define SQRT20 4.47213595499958
+
+double cksup[][2] = { {-DBL_MAX, DBL_MAX}, {-DBL_MAX, DBL_MAX}, {-DBL_MAX, DBL_MAX}, {-DBL_MAX, DBL_MAX}, 
+		      {-SQRT5, SQRT5}, {-SQRT5, SQRT5}, {-SQRT5, SQRT5}, {-SQRT5, SQRT5},
+		      {-1.0, 1.0},
+		      {-DBL_MAX, DBL_MAX}, {-DBL_MAX, DBL_MAX}, {-DBL_MAX, DBL_MAX}, {-DBL_MAX, DBL_MAX},
+		      {-SQRT20, SQRT20},{-SQRT20, SQRT20},{-SQRT20, SQRT20},{-SQRT20, SQRT20},
+		      {-DBL_MAX, DBL_MAX}, {-DBL_MAX, DBL_MAX}, {-DBL_MAX, DBL_MAX}, {-DBL_MAX, DBL_MAX},
+		      {-SQRT5, SQRT5}, {-SQRT5, SQRT5}, {-SQRT5, SQRT5}, {-SQRT5, SQRT5},
+		      {-0.0, 0.0}, // PLEASE DON'T EVER USE THIS
+		      {-DBL_MAX, DBL_MAX}, {-DBL_MAX, DBL_MAX}, {-DBL_MAX, DBL_MAX}, {-DBL_MAX, DBL_MAX}, 
+		      {-SQRT5, DBL_MAX}, {-SQRT5, DBL_MAX}, {-SQRT5, DBL_MAX}, {-SQRT5, DBL_MAX}, 
+		      {-1.0, DBL_MAX} };
 
 /* 
    np_kernelv does weighted products of vectors - this is useful for 
@@ -796,7 +823,8 @@ void np_ckernelv(const int KERNEL,
                  const double * const xt, const int num_xt, 
                  const int do_xw,
                  const double x, const double h, 
-                 double * const result){
+                 double * const result,
+		 const NL * const nl){
 
   /* 
      this should be read as:
@@ -821,8 +849,17 @@ void np_ckernelv(const int KERNEL,
                                    np_cdf_epan2, np_cdf_epan4, np_cdf_epan6, np_cdf_epan8, 
                                    np_cdf_rect };
 
-  for (i = 0, j = 0; i < num_xt; i++, j += bin_do_xw)
-    result[i] = xw[j]*k[KERNEL]((x-xt[i])/h);
+  if(nl == NULL)
+    for (i = 0, j = 0; i < num_xt; i++, j += bin_do_xw)
+      result[i] = xw[j]*k[KERNEL]((x-xt[i])/h);
+  else{
+    for (int m = 0; m < nl->n; m++){
+      const int istart = kdt_extern->kdn[nl->node[m]].istart;
+      const int nlev = kdt_extern->kdn[nl->node[m]].nlev;
+      for (i = istart, j = bin_do_xw*istart; i < istart+nlev; i++, j += bin_do_xw)
+	result[i] = xw[j]*k[KERNEL]((x-xt[i])/h);
+    }
+  }
 
 }
 
@@ -849,7 +886,8 @@ void np_ukernelv(const int KERNEL,
                  const double * const xt, const int num_xt, 
                  const int do_xw,
                  const double x, const double lambda, const int ncat,
-                 double * const result){
+                 double * const result,
+		 const NL * const nl){
 
   /* 
      this should be read as:
@@ -865,9 +903,17 @@ void np_ukernelv(const int KERNEL,
   double (* const k[])(int, double, int) = { np_uaa, np_uli_racine,
                                              np_econvol_uaa, np_econvol_uli_racine };
 
-  for (i = 0, j = 0; i < num_xt; i++, j += bin_do_xw)
-    result[i] = xw[j]*k[KERNEL]((xt[i]==x), lambda, ncat);
-
+  if(nl == NULL)
+    for (i = 0, j = 0; i < num_xt; i++, j += bin_do_xw)
+      result[i] = xw[j]*k[KERNEL]((xt[i]==x), lambda, ncat);
+  else{
+    for (int m = 0; m < nl->n; m++){
+      const int istart = kdt_extern->kdn[nl->node[m]].istart;
+      const int nlev = kdt_extern->kdn[nl->node[m]].nlev;
+      for (i = istart, j = bin_do_xw*istart; i < istart+nlev; i++, j += bin_do_xw)
+	result[i] = xw[j]*k[KERNEL]((xt[i]==x), lambda, ncat);
+    }
+  }
 }
 
 void np_convol_okernelv(const int KERNEL, 
@@ -890,7 +936,8 @@ void np_okernelv(const int KERNEL,
                  const double * const xt, const int num_xt, 
                  const int do_xw,
                  const double x, const double lambda,
-                 double * const result){
+                 double * const result,
+		 const NL * const nl){
   
   /* 
      this should be read as:
@@ -905,8 +952,17 @@ void np_okernelv(const int KERNEL,
 
   double (* const k[])(double, double, double) = { np_owang_van_ryzin, np_oli_racine };
 
-  for (i = 0, j = 0; i < num_xt; i++, j += bin_do_xw)
-    result[i] = xw[j]*k[KERNEL](xt[i], x, lambda);
+  if(nl == NULL)
+    for (i = 0, j = 0; i < num_xt; i++, j += bin_do_xw)
+      result[i] = xw[j]*k[KERNEL](xt[i], x, lambda);
+  else{
+    for (int m = 0; m < nl->n; m++){
+      const int istart = kdt_extern->kdn[nl->node[m]].istart;
+      const int nlev = kdt_extern->kdn[nl->node[m]].nlev;
+      for (i = istart, j = bin_do_xw*istart; i < istart+nlev; i++, j += bin_do_xw)
+	result[i] = xw[j]*k[KERNEL](xt[i], x, lambda);
+    }
+  }
 }
 
 // W = A
@@ -921,7 +977,8 @@ void np_outer_weighted_sum(double * const * const mat_A, double * const sgn_A, c
                            const int symmetric,
                            const int gather_scatter,
                            const int bandwidth_divide, const double dband,
-                           double * const result){
+                           double * const result,
+			   const NL * const nl){
 
   int i,j,k, l = parallel_sum?which_k:0;
   const int kstride = (parallel_sum ? (MAX(ncol_A, 1)*MAX(ncol_B, 1)) : 0);
@@ -952,90 +1009,218 @@ void np_outer_weighted_sum(double * const * const mat_A, double * const sgn_A, c
     temp = weights[which_k];
     weights[which_k] = 0.0;
   }
+  
+  if(nl == NULL){
+    if(!gather_scatter){
+      if(!symmetric){
+	if(kpow == 1){
+	  for (k = 0; k < num_weights; k++, l+=linc)
+	    for (j = 0; j < max_A; j++)
+	      for (i = 0; i < max_B; i++)
+		result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*weights[k]/db;
+	} else { // kpow != 1
+	  for (k = 0; k < num_weights; k++, l+=linc)
+	    for (j = 0; j < max_A; j++)
+	      for (i = 0; i < max_B; i++)
+		result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*ipow(weights[k]/db, kpow);
+	}
+      } else { // symmetric
+	if(kpow == 1){
+	  for (k = 0; k < num_weights; k++, l+=linc){
+	    for (j = 0; j < max_A; j++){
+	      for (i = 0; i <= j; i++){
+		result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*weights[k]/db;
+	      }
+	    }
+	  }
+	} else { // kpow != 1
+	  for (k = 0; k < num_weights; k++, l+=linc){
+	    for (j = 0; j < max_A; j++){
+	      for (i = 0; i <= j; i++){
+		result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*ipow(weights[k]/db, kpow);
+	      }
+	    }
+	  }
+	}
 
-  if(!gather_scatter){
-    if(!symmetric){
-      if(kpow == 1){
-        for (k = 0; k < num_weights; k++, l+=linc)
-          for (j = 0; j < max_A; j++)
-            for (i = 0; i < max_B; i++)
-              result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*weights[k]/db;
-      } else { // kpow != 1
-        for (k = 0; k < num_weights; k++, l+=linc)
-          for (j = 0; j < max_A; j++)
-            for (i = 0; i < max_B; i++)
-              result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*ipow(weights[k]/db, kpow);
-      }
-    } else { // symmetric
-      if(kpow == 1){
-        for (k = 0; k < num_weights; k++, l+=linc){
-          for (j = 0; j < max_A; j++){
-            for (i = 0; i <= j; i++){
-              result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*weights[k]/db;
-            }
-          }
-        }
-      } else { // kpow != 1
-        for (k = 0; k < num_weights; k++, l+=linc){
-          for (j = 0; j < max_A; j++){
-            for (i = 0; i <= j; i++){
-              result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*ipow(weights[k]/db, kpow);
-            }
-          }
-        }
+
+	for (j = 0; j < max_A; j++){
+	  for (i = (max_A-1); i > j; i--){
+	    result[j*max_B+i] = result[i*max_B+j];
+	  }
+	}
       }
+    } else { // gather_scatter
+      if(!symmetric){
+	if(kpow == 1){
+	  for (k = 0; k < num_weights; k++, l+=linc)
+	    for (j = 0; j < max_A; j++)
+	      for (i = 0; i < max_B; i++)
+		result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*weights[k]/db;
+	} else { // kpow != 1
+	  for (k = 0; k < num_weights; k++, l+=linc)
+	    for (j = 0; j < max_A; j++)
+	      for (i = 0; i < max_B; i++)
+		result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*ipow(weights[k]/db, kpow);
+	}
+      } else { // symmetric
+	if(kpow == 1){
+	  for (k = 0; k < num_weights; k++){
+	    for (j = 0; j < max_A; j++){
+	      for (i = 0; i <= j; i++){
+		const double tp = pmat_A[j][k*have_A]*pmat_B[i][k*have_B]*weights[k]/db;
+		const double sgnp = tp*p_sgn[j*have_sgn]*p_sgn[i*have_sgn];
+		result[k*kstride+j*max_B+i] += tp;
+		result[(k+1)*max_AB+j*max_B+i] += sgnp;
+		//fprintf(stderr,"\nmax ab %d\tsgnp %e\tres %e",max_AB,sgnp,result[(k+1)*max_AB+j*max_B+i]);
+	      }
+	    }
+	    // insert reference weight in top right corner
+	    result[(k+1)*max_AB+max_B-1] = weights[k]/db;
+	  }
+	} else { // kpow != 1
+	  for (k = 0; k < num_weights; k++){
+	    for (j = 0; j < max_A; j++){
+	      for (i = 0; i <= j; i++){
+		result[k*kstride+j*max_B+i] += pmat_A[j][k*have_A]*pmat_B[i][k*have_B]*ipow(weights[k]/db, kpow);
+	      }
+	    }
+	  }
+	}
 
+	for (j = 0; j < max_A; j++){
+	  for (i = (max_A-1); i > j; i--){
+	    result[j*max_B+i] = result[i*max_B+j];
+	  }
+	}
 
-      for (j = 0; j < max_A; j++){
-        for (i = (max_A-1); i > j; i--){
-          result[j*max_B+i] = result[i*max_B+j];
-        }
       }
     }
-  } else { // gather_scatter
-    if(!symmetric){
-      if(kpow == 1){
-        for (k = 0; k < num_weights; k++, l+=linc)
-          for (j = 0; j < max_A; j++)
-            for (i = 0; i < max_B; i++)
-              result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*weights[k]/db;
-      } else { // kpow != 1
-        for (k = 0; k < num_weights; k++, l+=linc)
-          for (j = 0; j < max_A; j++)
-            for (i = 0; i < max_B; i++)
-              result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*ipow(weights[k]/db, kpow);
-      }
-    } else { // symmetric
-      if(kpow == 1){
-        for (k = 0; k < num_weights; k++){
-          for (j = 0; j < max_A; j++){
-            for (i = 0; i <= j; i++){
-              const double tp = pmat_A[j][k*have_A]*pmat_B[i][k*have_B]*weights[k]/db;
-              const double sgnp = tp*p_sgn[j*have_sgn]*p_sgn[i*have_sgn];
-              result[k*kstride+j*max_B+i] += tp;
-              result[(k+1)*max_AB+j*max_B+i] += sgnp;
-              //fprintf(stderr,"\nmax ab %d\tsgnp %e\tres %e",max_AB,sgnp,result[(k+1)*max_AB+j*max_B+i]);
-            }
-          }
-          // insert reference weight in top right corner
-          result[(k+1)*max_AB+max_B-1] = weights[k]/db;
-        }
-      } else { // kpow != 1
-        for (k = 0; k < num_weights; k++){
-          for (j = 0; j < max_A; j++){
-            for (i = 0; i <= j; i++){
-              result[k*kstride+j*max_B+i] += pmat_A[j][k*have_A]*pmat_B[i][k*have_B]*ipow(weights[k]/db, kpow);
-            }
-          }
-        }
-      }
+  } else { // TREE SUPPORT
+    if(!gather_scatter){
+      if(!symmetric){
+	if(kpow == 1){
+	  for (int m = 0; m < nl->n; m++){
+	    const int istart = kdt_extern->kdn[nl->node[m]].istart;
+	    const int nlev = kdt_extern->kdn[nl->node[m]].nlev;
+	    l = linc ? istart : l;
+	    for (k = istart; k < istart+nlev; k++, l+=linc)
+	      for (j = 0; j < max_A; j++)
+		for (i = 0; i < max_B; i++)
+		  result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*weights[k]/db;
+	  }
+	} else { // kpow != 1
+	  for (int m = 0; m < nl->n; m++){
+	    const int istart = kdt_extern->kdn[nl->node[m]].istart;
+	    const int nlev = kdt_extern->kdn[nl->node[m]].nlev;
+	    l = linc ? istart : l;
+	    for (k = istart; k < istart+nlev; k++, l+=linc)
+	      for (j = 0; j < max_A; j++)
+		for (i = 0; i < max_B; i++)
+		  result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*ipow(weights[k]/db, kpow);
+	  }
+	}
+      } else { // symmetric
+	if(kpow == 1){
+	  for (int m = 0; m < nl->n; m++){
+	    const int istart = kdt_extern->kdn[nl->node[m]].istart;
+	    const int nlev = kdt_extern->kdn[nl->node[m]].nlev;
+	    l = linc ? istart : l;
+	    for (k = istart; k < istart+nlev; k++, l+=linc){
+	      for (j = 0; j < max_A; j++){
+		for (i = 0; i <= j; i++){
+		  result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*weights[k]/db;
+		}
+	      }
+	    }
+	  }
+	} else { // kpow != 1
+	  for (int m = 0; m < nl->n; m++){
+	    const int istart = kdt_extern->kdn[nl->node[m]].istart;
+	    const int nlev = kdt_extern->kdn[nl->node[m]].nlev;
+	    l = linc ? istart : l;
+	    for (k = istart; k < istart+nlev; k++, l+=linc){
+	      for (j = 0; j < max_A; j++){
+		for (i = 0; i <= j; i++){
+		  result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*ipow(weights[k]/db, kpow);
+		}
+	      }
+	    }
+	  }
+	}
 
-      for (j = 0; j < max_A; j++){
-        for (i = (max_A-1); i > j; i--){
-          result[j*max_B+i] = result[i*max_B+j];
-        }
-      }
 
+	for (j = 0; j < max_A; j++){
+	  for (i = (max_A-1); i > j; i--){
+	    result[j*max_B+i] = result[i*max_B+j];
+	  }
+	}
+      }
+    } else { // gather_scatter
+      if(!symmetric){
+	if(kpow == 1){
+	  for (int m = 0; m < nl->n; m++){
+	    const int istart = kdt_extern->kdn[nl->node[m]].istart;
+	    const int nlev = kdt_extern->kdn[nl->node[m]].nlev;
+	    l = linc ? istart : l;
+	    for (k = istart; k < istart+nlev; k++, l+=linc)
+	      for (j = 0; j < max_A; j++)
+		for (i = 0; i < max_B; i++)
+		  result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*weights[k]/db;
+	  }
+	} else { // kpow != 1
+	  for (int m = 0; m < nl->n; m++){
+	    const int istart = kdt_extern->kdn[nl->node[m]].istart;
+	    const int nlev = kdt_extern->kdn[nl->node[m]].nlev;
+	    l = linc ? istart : l;
+	    for (k = istart; k < istart+nlev; k++, l+=linc)
+	      for (j = 0; j < max_A; j++)
+		for (i = 0; i < max_B; i++)
+		  result[k*kstride+j*max_B+i] += pmat_A[j][l*have_A]*pmat_B[i][l*have_B]*ipow(weights[k]/db, kpow);
+	  }
+	}
+      } else { // symmetric
+	if(kpow == 1){
+	  for (int m = 0; m < nl->n; m++){
+	    const int istart = kdt_extern->kdn[nl->node[m]].istart;
+	    const int nlev = kdt_extern->kdn[nl->node[m]].nlev;
+	    l = linc ? istart : l;
+	    for (k = istart; k < istart+nlev; k++){
+	      for (j = 0; j < max_A; j++){
+		for (i = 0; i <= j; i++){
+		  const double tp = pmat_A[j][k*have_A]*pmat_B[i][k*have_B]*weights[k]/db;
+		  const double sgnp = tp*p_sgn[j*have_sgn]*p_sgn[i*have_sgn];
+		  result[k*kstride+j*max_B+i] += tp;
+		  result[(k+1)*max_AB+j*max_B+i] += sgnp;
+		  //fprintf(stderr,"\nmax ab %d\tsgnp %e\tres %e",max_AB,sgnp,result[(k+1)*max_AB+j*max_B+i]);
+		}
+	      }
+	      // insert reference weight in top right corner
+	      result[(k+1)*max_AB+max_B-1] = weights[k]/db;
+	    }
+	  }
+	} else { // kpow != 1
+	  for (int m = 0; m < nl->n; m++){
+	    const int istart = kdt_extern->kdn[nl->node[m]].istart;
+	    const int nlev = kdt_extern->kdn[nl->node[m]].nlev;
+	    l = linc ? istart : l;
+	    for (k = istart; k < istart+nlev; k++){
+	      for (j = 0; j < max_A; j++){
+		for (i = 0; i <= j; i++){
+		  result[k*kstride+j*max_B+i] += pmat_A[j][k*have_A]*pmat_B[i][k*have_B]*ipow(weights[k]/db, kpow);
+		}
+	      }
+	    }
+	  }
+	}
+
+	for (j = 0; j < max_A; j++){
+	  for (i = (max_A-1); i > j; i--){
+	    result[j*max_B+i] = result[i*max_B+j];
+	  }
+	}
+
+      }
     }
   }
 
@@ -1087,7 +1272,23 @@ double *weighted_sum){
   /* Declarations */
 
   int i,j,l, mstep, js, je, num_obs_eval_alloc, sum_element_length;
-  int do_psum; 
+  int do_psum, switch_te; 
+
+  int np_ks_tree_use = (int_TREE == NP_TREE_TRUE) && (operator != OP_CONVOLUTION) && (BANDWIDTH_reg == BW_FIXED);
+
+  NL nl = {.node = NULL, .n = 0, .nalloc = 0};
+  NL * pnl=  np_ks_tree_use ? &nl : NULL;
+
+#ifdef MPI2
+  // switch parallelisation strategies based on biggest stride
+ 
+  int stride = MAX(ceil((double) num_obs_eval / (double) iNum_Processors),1);
+
+  num_obs_eval_alloc = stride*iNum_Processors;
+
+#else
+  num_obs_eval_alloc = num_obs_eval;
+#endif
 
   const int KERNEL_reg_np = KERNEL_reg + OP_CFUN_OFFSETS[operator];
   const int KERNEL_unordered_reg_np = KERNEL_unordered_reg + OP_UFUN_OFFSETS[operator];
@@ -1096,7 +1297,7 @@ double *weighted_sum){
     (MAX(num_var_continuous_extern, 1) * MAX(num_var_ordered_extern, 1));
 
   double *lambda, **matrix_bandwidth, **matrix_eval_bandwidth = NULL, *m = NULL;
-  double *tprod, dband, *buf, *ws;
+  double *tprod, dband, *ws;
 
   double * const * const xtc = (BANDWIDTH_reg == BW_ADAP_NN)?
     matrix_X_continuous_eval:matrix_X_continuous_train;
@@ -1112,13 +1313,6 @@ double *weighted_sum){
   double * const * const xo = (BANDWIDTH_reg == BW_ADAP_NN)?
     matrix_X_ordered_train:matrix_X_ordered_eval;
 
-#ifdef MPI2
-  int stride = MAX(ceil((double) num_obs_eval / (double) iNum_Processors),1);
-  num_obs_eval_alloc = stride*iNum_Processors;
-#else
-  num_obs_eval_alloc = num_obs_eval;
-#endif
-
   if (num_obs_eval == 0) {
     return(1);
   }
@@ -1137,20 +1331,6 @@ double *weighted_sum){
   sum_element_length = MAX(num_var_continuous_extern, 1) * 
     MAX(num_var_ordered_extern, 1);
 
-  buf = alloc_vecd(num_obs_eval_alloc*sum_element_length);
-
-  /* clear memory in temporary buffer and in result */
-  if(gather_scatter){
-    for(i = 0; i < num_obs_eval_alloc*sum_element_length; i++){
-      buf[i] = weighted_sum[i];
-    }
-  }else{
-    for(i = 0; i < num_obs_eval_alloc*sum_element_length; i++){
-      buf[i] = 0.0;
-      weighted_sum[i] = 0.0;
-    }    
-  }
-
   /* assert(!(BANDWIDTH_reg == BW_ADAP_NN)); */
   /* Conduct the estimation */
 
@@ -1225,21 +1405,40 @@ double *weighted_sum){
     return(1);
   }
 
+  if(!gather_scatter)
+    for(i = 0; i < num_obs_eval_alloc*sum_element_length; i++){
+      weighted_sum[i] = 0.0;
+    }
+
   if (BANDWIDTH_reg == BW_FIXED || BANDWIDTH_reg == BW_GEN_NN){
 #ifdef MPI2
-    js = stride * my_rank;
-    je = MIN(num_obs_eval - 1, js + stride - 1);
+    if(!gather_scatter){
+      js = stride * my_rank;
+      je = MIN(num_obs_eval - 1, js + stride - 1);
+      ws = weighted_sum + js*sum_element_length;
+    } else {
+      js = 0;
+      je = num_obs_eval - 1;
+      ws = weighted_sum;
+    }
 #else
     js = 0;
     je = num_obs_eval - 1;
+    ws = weighted_sum;
 #endif
     
-    ws = weighted_sum + js * sum_element_length;
+
   } else {
 #ifdef MPI2
-    js = stride * my_rank;
-    je = MIN(num_obs_train - 1, js + stride - 1);
-    ws = buf;
+    if(!gather_scatter){
+      js = stride * my_rank;
+      je = MIN(num_obs_train - 1, js + stride - 1);
+      ws = weighted_sum;
+    } else {
+      js = 0;
+      je = num_obs_train - 1;
+      ws = weighted_sum;
+    }
 #else
     js = 0;
     je = num_obs_train - 1;
@@ -1261,13 +1460,32 @@ double *weighted_sum){
 
     l = 0;
 
+    // do a hail mary, then generate the interaction list
+    // anything but a fixed bandwidth is not yet supported
+    // that includes convolutions 
+
+    if(np_ks_tree_use){
+      // reset the interaction node list
+      nl.n = 0;
+      double bb[kdt_extern->ndim*2];
+
+      for(i = 0; i < num_reg_continuous; i++){
+	bb[2*i] = -cksup[KERNEL_reg_np][1];
+	bb[2*i+1] = -cksup[KERNEL_reg_np][0];
+
+	bb[2*i] = (fabs(bb[2*i]) == DBL_MAX) ? bb[2*i] : (xc[i][j] + m[i]*bb[2*i]);
+	bb[2*i+1] = (fabs(bb[2*i+1]) == DBL_MAX) ? bb[2*i+1] : (xc[i][j] + m[i]*bb[2*i+1]);
+      }
+
+      boxSearch(kdt_extern, 0, bb, pnl);
+    }
     /* continuous first */
 
     /* for the first iteration, no weights */
     /* for the rest, the accumulated products are the weights */
     if((BANDWIDTH_reg != BW_ADAP_NN) || (operator != OP_CONVOLUTION)){
       for(i=0; i < num_reg_continuous; i++, l++, m += mstep){
-        np_ckernelv(KERNEL_reg_np, xtc[i], num_xt, l, xc[i][j], *m, tprod);
+        np_ckernelv(KERNEL_reg_np, xtc[i], num_xt, l, xc[i][j], *m, tprod, pnl);
         dband *= *m;
       }
     } else {
@@ -1282,7 +1500,7 @@ double *weighted_sum){
 
     for(i=0; i < num_reg_unordered; i++, l++){
       np_ukernelv(KERNEL_unordered_reg_np, xtu[i], num_xt, l, xu[i][j], 
-                  lambda[i], num_categories[i], tprod);
+                  lambda[i], num_categories[i], tprod, pnl);
     }
 
     /* ordered third */
@@ -1290,7 +1508,7 @@ double *weighted_sum){
       for(i=0; i < num_reg_ordered; i++, l++){
         np_okernelv(KERNEL_ordered_reg, xto[i], num_xt, l,
                     xo[i][j], lambda[num_reg_unordered+i], 
-                    tprod);
+                    tprod, pnl);
       }
     } else {
       for(i=0; i < num_reg_ordered; i++, l++){
@@ -1313,17 +1531,24 @@ double *weighted_sum){
                           symmetric,
                           gather_scatter,
                           bandwidth_divide, dband,
-                          ws);
+                          ws, pnl);
 
   }
 
+  if(!gather_scatter){ 
+    // gather_scatter is only used for the local-linear cv
+    // note: ll cv + adaptive_nn does not work in parallel
 #ifdef MPI2
-  if (BANDWIDTH_reg == BW_FIXED || BANDWIDTH_reg == BW_GEN_NN){
-    MPI_Allgather(weighted_sum + js * sum_element_length, stride * sum_element_length, MPI_DOUBLE, weighted_sum, stride * sum_element_length, MPI_DOUBLE, comm[1]);
-  } else if(BANDWIDTH_reg == BW_ADAP_NN){
-    MPI_Allreduce(ws, weighted_sum, num_obs_eval*sum_element_length, MPI_DOUBLE, MPI_SUM, comm[1]);
-  }
+    if (BANDWIDTH_reg == BW_FIXED || BANDWIDTH_reg == BW_GEN_NN){
+      MPI_Allgather(MPI_IN_PLACE, stride * sum_element_length, MPI_DOUBLE, weighted_sum, stride * sum_element_length, MPI_DOUBLE, comm[1]);
+    } else if(BANDWIDTH_reg == BW_ADAP_NN){
+      MPI_Allreduce(MPI_IN_PLACE, weighted_sum, num_obs_eval*sum_element_length, MPI_DOUBLE, MPI_SUM, comm[1]);
+    }
 #endif
+  }
+
+  if(nl.node != NULL)
+    free(nl.node);
 
   free(lambda);
   free_tmat(matrix_bandwidth);
@@ -1332,7 +1557,6 @@ double *weighted_sum){
     free_tmat(matrix_eval_bandwidth);
 
   free(tprod);
-  free(buf);
   
   return(0);
 }
@@ -1699,17 +1923,27 @@ double *vector_scale_factor,
 int *num_categories){
 
   // note that mean has 2*num_obs allocated for npksum
-  int i, j, l, sf_flag = 0;
+  int i, j, l, sf_flag = 0, num_obs_eval_alloc, tsf;
 
   double cv = 0.0;
   double * lambda = NULL, * vsf = NULL;
   double ** matrix_bandwidth = NULL;
 
-  double aicc = 0.0;
+  double  * aicc;
   double traceH = 0.0;
 
   const int leave_one_out = (bwm == RBWM_CVLS)?1:0;
 
+#ifdef MPI2
+    int stride = MAX(ceil((double) num_obs / (double) iNum_Processors),1);
+    num_obs_eval_alloc = stride*iNum_Processors;
+    aicc = (double *)malloc(stride*sizeof(double));
+#else
+    num_obs_eval_alloc = num_obs;
+    aicc = (double *)malloc(sizeof(double));
+#endif
+
+
   // Allocate memory for objects 
 
   lambda = alloc_vecd(num_reg_unordered+num_reg_ordered);
@@ -1747,6 +1981,11 @@ int *num_categories){
     num_var_continuous_extern = 0; // rows in the y_mat
     num_var_ordered_extern = 0; // rows in weights
 
+    // workaround for bwscaling = TRUE
+    // really just want to get the full product kernel evaluated at zero
+    tsf = int_LARGE_SF;
+
+    int_LARGE_SF = 1;
     kernel_weighted_sum_np(KERNEL_reg,
                            KERNEL_unordered_reg,
                            KERNEL_ordered_reg,
@@ -1775,7 +2014,11 @@ int *num_categories){
                            vector_scale_factor,
                            num_categories,
                            NULL,
-                           &aicc);
+                           aicc);
+    int_LARGE_SF = tsf;
+    num_var_continuous_extern = 0; 
+
+
     //fprintf(stderr,"\n%e\n",aicc);
   }
 
@@ -1786,7 +2029,7 @@ int *num_categories){
     // Generate bandwidth vector given scale factors, nearest neighbors, or lambda 
 
     double * lc_Y[2];
-    double * mean = (double *)malloc(2*num_obs*sizeof(double));
+    double * mean = (double *)malloc(2*num_obs_eval_alloc*sizeof(double));
 
     num_var_continuous_extern = 2; // columns in the y_mat
     num_var_ordered_extern = 0; // columns in weights
@@ -1825,6 +2068,9 @@ int *num_categories){
                            num_categories,
                            NULL,
                            mean);
+
+    num_var_continuous_extern = 0;
+
     
     // every even entry in mean is sum(y*kij)
     // every odd is sum(kij)
@@ -1835,7 +2081,7 @@ int *num_categories){
       const double dy = vector_Y[ii]-mean[ii2]/sk;
       cv += dy*dy;
       if(bwm == RBWM_CVAIC)
-        traceH += aicc/sk;
+        traceH += aicc[0]/sk;
 
       //fprintf(stderr,"mj: %e\n",mean[ii2]/(MAX(DBL_MIN, mean[ii2+1])));
     }
@@ -1925,7 +2171,7 @@ int *num_categories){
     }
 
 
-    for(j = 0; j < num_obs; j++){
+    for(j = 0; j < num_obs; j++){ // main loop
       nepsilon = 0.0;
 
       for(l = 0; l < (nrc1); l++){
@@ -1933,6 +2179,102 @@ int *num_categories){
         XTKY[l] = &kwm[j*nrcc22+l+1];
       }
 
+#ifdef MPI2
+      if((j % iNum_Processors) == 0){
+
+        // some guys have to sit out the last calculation, but
+        // they still sync up afterwards
+        if((j+my_rank) < (num_obs-1)){
+
+          for(l = 0; l < (nrc2); l++){
+            XTKX[l] = PXTKX[l] + j + my_rank + 1;
+          }
+
+          for(l = 0; l < num_reg_continuous; l++)
+            PXC[l] = matrix_X_continuous[l] + j + my_rank + 1;
+
+          for(l = 0; l < num_reg_unordered; l++)
+            PXU[l] = matrix_X_unordered[l] + j + my_rank + 1;
+
+          for(l = 0; l < num_reg_ordered; l++)
+            PXO[l] = matrix_X_ordered[l] + j + my_rank + 1;
+        
+          for(l = 0; l < num_reg_continuous; l++){
+        
+            for(i = 0; i < (num_obs-j-1-my_rank); i++){
+              XTKX[l+2][i] = matrix_X_continuous[l][i+j+1+my_rank]-matrix_X_continuous[l][j+my_rank];
+            }
+            TCON[l][0] = matrix_X_continuous[l][j+my_rank]; // temporary storage
+          }
+
+          for(l = 0; l < num_reg_unordered; l++)
+            TUNO[l][0] = matrix_X_unordered[l][j+my_rank];
+
+          for(l = 0; l < num_reg_ordered; l++)
+            TORD[l][0] = matrix_X_ordered[l][j+my_rank];
+
+          num_var_continuous_extern = nrc2; // rows in the y_mat
+          num_var_ordered_extern = nrc2; // rows in weights
+
+          kernel_weighted_sum_np(KERNEL_reg,
+                                 KERNEL_unordered_reg,
+                                 KERNEL_ordered_reg,
+                                 BANDWIDTH_reg,
+                                 num_obs-j-my_rank-1,
+                                 1,
+                                 num_reg_unordered,
+                                 num_reg_ordered,
+                                 num_reg_continuous,
+                                 0, // we leave one out via the weight matrix
+                                 1, // kernel_pow = 1
+                                 (BANDWIDTH_reg == BW_ADAP_NN)?1:0, // bandwidth_divide = FALSE when not adaptive
+                                 0, // do_smooth_coef_weights = FALSE (not implemented)
+                                 1, // symmetric
+                                 1, // gather-scatter sum
+                                 0, // no convolution
+                                 PXU, // TRAIN
+                                 PXO, 
+                                 PXC,
+                                 TUNO, // EVAL
+                                 TORD,
+                                 TCON,
+                                 XTKX,
+                                 XTKX,
+                                 sgn,
+                                 vsf,
+                                 num_categories,
+                                 NULL,
+                                 kwm+(j+my_rank)*nrcc22); // weighted sum
+
+          num_var_continuous_extern = 0; // set back to number of regressors
+          num_var_ordered_extern = 0; // always zero
+
+          // need to use reference weight to fix weight sum
+          for(int jj = j+my_rank+1; jj < num_obs; jj++){
+            const double RW = kwm[jj*nrcc22+nrc1]*(XTKX[0][-1]-XTKX[0][jj-j-my_rank-1]);
+            for(int ii = 1; ii < nrc2; ii++){
+              kwm[jj*nrcc22+ii*nrc2] += RW*XTKX[ii][jj-j-my_rank-1]*sgn[ii];
+            }
+          }
+        }
+        // reduce all work arrays
+        const int nrem = num_obs % iNum_Processors;
+        const int nred = ((j+iNum_Processors) > num_obs) ? nrem : iNum_Processors;
+
+        MPI_Allreduce(MPI_IN_PLACE, kwm+j*nrcc22, nred*nrcc22, MPI_DOUBLE, MPI_SUM, comm[1]);
+      }
+
+      // due to a quirk of the algorithm in parallel, always need to re-symmetrise array
+
+      double * const tpk = kwm+j*nrcc22;
+      for (int jj = 0; jj < (nrc2); jj++){
+        for (int ii = (nrc1); ii > jj; ii--){
+          tpk[jj*(nrc2)+ii] = tpk[ii*(nrc2)+jj];
+        }
+      }
+
+#else
+
       if(j < (num_obs-1)){
 
         for(l = 0; l < (nrc2); l++){
@@ -1963,7 +2305,6 @@ int *num_categories){
         for(l = 0; l < num_reg_ordered; l++)
           TORD[l][0] = matrix_X_ordered[l][j];
       
-
         num_var_continuous_extern = nrc2; // rows in the y_mat
         num_var_ordered_extern = nrc2; // rows in weights
 
@@ -1996,6 +2337,10 @@ int *num_categories){
                                num_categories,
                                NULL,
                                kwm+j*nrcc22); // weighted sum
+
+        num_var_continuous_extern = 0; // set back to number of regressors
+        num_var_ordered_extern = 0; // always zero
+
         // need to use reference weight to fix weight sum
         for(int jj = j+1; jj < num_obs; jj++){
           const double RW = kwm[jj*nrcc22+nrc1]*(XTKX[0][-1]-XTKX[0][jj-j-1]);
@@ -2011,12 +2356,13 @@ int *num_categories){
           }
         }
       }
+#endif
       
       // need to manipulate KWM pointers and XTKY - done
 
       if(bwm == RBWM_CVAIC){
-        KWM[0][0] += aicc;
-        XTKY[0][0] += aicc*vector_Y[j];
+        KWM[0][0] += aicc[0];
+        XTKY[0][0] += aicc[0]*vector_Y[j];
       }
 
       while(mat_inv(KWM, XTKXINV) == NULL){ // singular = ridge about
@@ -2026,7 +2372,7 @@ int *num_categories){
       }
       
       if(bwm == RBWM_CVAIC)
-        traceH += XTKXINV[0][0]*aicc;
+        traceH += XTKXINV[0][0]*aicc[0];
    
       XTKY[0][0] += nepsilon*XTKY[0][0]/NZD(KWM[0][0]);
 
@@ -2063,12 +2409,19 @@ int *num_categories){
     free(sgn);
   }
 
+  free(aicc);
   free(lambda);
   free_mat(matrix_bandwidth,num_reg_continuous);
 
+	/* Negative penalties are treated as infinite: Hurvich et al pg 277 */
+
   cv /= (double)num_obs;
   if(bwm == RBWM_CVAIC){
-    cv = log(cv) + (1.0+traceH/((double)num_obs))/(1.0-(traceH+2.0)/((double)num_obs));
+    if((1.0+traceH/((double)num_obs))/(1.0-(traceH+2.0)/((double)num_obs)) < 0) {
+      cv = DBL_MAX;
+    } else {
+      cv = log(cv) + (1.0+traceH/((double)num_obs))/(1.0-(traceH+2.0)/((double)num_obs));
+    }
   }
   return(cv);
 }
